#ifndef EXPRESSION_H
#define EXPRESSION_H

#include <memory>
#include <string>
#include "types.hpp"
#include "symbol-table.hpp"

class Expression;
typedef std::shared_ptr<Expression> ExpressionPtr;

class Expression {
   public:
      virtual ~Expression() {};
      virtual Value evaluate() const = 0;
};

class Binary: public Expression {
   public:
      typedef Value (*BinaryOp)(Value val1, Value val2);
      Binary(BinaryOp op_, ExpressionPtr expr1_, ExpressionPtr expr2_) :
	 op{op_}, expr1{expr1_}, expr2{expr2_} {
      }
      virtual Value evaluate() const {
	 return op(expr1->evaluate(), expr2->evaluate());
      }
   private:
      BinaryOp op;
      ExpressionPtr expr1;
      ExpressionPtr expr2;
};

class Primary: public Expression {
   public:
      Primary(Value value_) : value{value_} {};
      virtual Value evaluate() const {
	 return value;
      }
   private:
      Value value;
};

class Variable: public Expression {
   public:
      Variable(SymbolTable& symtab_, const std::string& varname_) :
	 symtab{symtab_}, varname{varname_} {
      }
      virtual Value evaluate() const {
	 ExpressionPtr expr = symtab.get(varname);
	 if (expr) {
	    return expr->evaluate();
	 } else {
	    return Value();
	 }
      }
      void set(ExpressionPtr expr) {
	 symtab.define(varname, expr);
      }
   private:
      SymbolTable& symtab;
      std::string varname;
};
typedef std::shared_ptr<Variable> VariablePtr;

#endif
