#ifndef LEX_H
#define LEX_H

#include <iostream>
#include "token.hpp"

class Lex {
   public:
      Lex(std::istream& in_, std::ostream& out_,
	 const std::string& prompt1_, const std::string& prompt2_) :
	    in{in_}, out{out_}, prompt1{prompt1_}, prompt2{prompt2_},
	    eof{false}, error{false}, ch{0}, start{true}, finished{true} {
	 next();
      }

      // accessors
      bool valid() const {
	 /* check if this lex object continues to read from
	    its input. Once this returns false, the return
	    value remains false fur all further invocations */
	 return !eof && !error;
      }

      // mutators
      void markFinished() {
	 /* tell the lexical analyser that a complete
	    expression has been read. This information is used
	    to print the correct prompt. */
	 finished = true;
      }

      Token getToken() {
	 /* read and return the next token from the input.
	    Note however that no further read operations are
	    performed once an error or end of file has been
	    found. */
	 while (!error && !eof && isspace(ch)) {
	    next();
	 }
	 if (error) {
	    return Token(Token::ERROR);
	 }
	 if (eof) {
	    return Token(Token::END);
	 }
	 finished = false;

	 // simple symbols
	 switch (ch) {
	    case '(':
	       next();
	       return Token(Token::LPAREN);

	    case ')':
	       next();
	       return Token(Token::RPAREN);

	    case '=':
	       next();
	       return Token(Token::BECOMES);

	    case '+':
	       next();
	       return Token(Token::PLUS);

	    case '-':
	       next();
	       return Token(Token::MINUS);

	    case '*':
	       next();
	       return Token(Token::MUL);

	    case '/':
	       next();
	       return Token(Token::DIV);

	    case ';':
	       next();
	       return Token(Token::SEMICOLON);
	 }

	 if (isalpha(ch)) {
	    // variable name
	    std::string ident; ident += ch;
	    next();
	    while (!eof && !error && (isalnum(ch) || ch == '-')) {
	       ident += ch; next();
	    }
	    return Token(Token::VARIABLE, ident);
	 }

	 if (isdigit(ch)) {
	    // constant value
	    Value val = ch - '0';
	    next();
	    while (!eof && !error && isdigit(ch)) {
	       val = val * 10 + ch - '0'; next();
	    }
	    return Token(Token::VALUE, val);
	 }

	 error = true;
	 return Token(Token::ERROR);
      }

      // istream-like behaviour
      operator bool() const {
	 return valid();
      }

      Lex& operator>>(Token& token) {
	 token = getToken(); return *this;
      }

   private:
      std::istream& in; // input stream
      std::ostream& out; // used for prompts
      const std::string prompt1; // at the beginning of an expression
      const std::string prompt2; // in the middle of an expression
      bool eof; // eof detected? this flag is sticky!
      bool error; // error detected? this is sticky as well
      char ch; // current character read
      bool start; // at the beginning of a line?
      bool finished; // was one expression finished?

      void next() {
	 if (eof || error) return;
	 if (start) {
	    if (finished) {
	       out << prompt1;
	    } else {
	       out << prompt2;
	    }
	    out.flush();
	    start = false;
	 }
	 ch = in.get();
	 if (in) {
	    if (ch == '\n') {
	       start = true;
	    }
	 } else {
	    ch = 0;
	    if (in.eof()) {
	       eof = true;
	    } else {
	       error = true;
	    }
	 }
      }
};

#endif
