#include <sstream>
#include "token.hpp"
#include "parser.hpp"

static Value addop(Value expr1, Value expr2) {
   return expr1 + expr2;
}

static Value subop(Value expr1, Value expr2) {
   return expr1 - expr2;
}

static Value mulop(Value expr1, Value expr2) {
   return expr1 * expr2;
}

static Value divop(Value expr1, Value expr2) {
   return expr1 / expr2;
}

Parser::Exception::Exception(const Token& token, const std::string& _msg) {
   std::ostringstream osmsg;
   osmsg << "At " << token << ": " << _msg;
   msg = osmsg.str();
}

Parser::Exception::~Exception() noexcept {};

const char* Parser::Exception::what() const noexcept {
   return msg.c_str();
}

Parser::Parser(Lex& _lex, SymbolTable& _symtab) :
      lex(_lex), symtab(_symtab), tokenConsumed(true) {
}

ExpressionPtr Parser::getExpr() throw(Exception) {
   getToken();
   if (!lex.valid()) {
      return ExpressionPtr((Expression*) 0);
   }
   ExpressionPtr expr = parseAssignment();
   if (getToken().symbol != Token::SEMICOLON) {
      throw Exception(getToken(), "';' expected");
   }
   tokenConsumed = true;
   lex.markFinished();
   return expr;
}

void Parser::nextToken() {
   if (tokenConsumed) {
      token = lex.getToken();
   } else {
      tokenConsumed = true;
   }
}

Token Parser::getToken() {
   if (tokenConsumed) {
      token = lex.getToken();
      tokenConsumed = false;
   }
   return token;
}

ExpressionPtr Parser::parseAssignment() throw(Exception) {
   ExpressionPtr expr = parseSimpleExpression();
   if (getToken().symbol == Token::BECOMES) {
      VariablePtr var = std::dynamic_pointer_cast<Variable>(expr);
      if (!var) {
	 throw Exception(getToken(), "variable expected");
      }
      nextToken();
      ExpressionPtr expr2 = parseSimpleExpression();
      var->set(expr2);
      return expr2;
   }
   return expr;
}

ExpressionPtr Parser::parseSimpleExpression() throw(Exception) {
   ExpressionPtr expr = parseTerm();
   while (getToken().symbol == Token::PLUS ||
	 getToken().symbol == Token::MINUS) {
      Binary::BinaryOp op;
      switch (getToken().symbol) {
	 case Token::PLUS: op = addop; break;
	 case Token::MINUS: op = subop; break;
	 default: /* does not happen */ break;
      }
      nextToken();
      ExpressionPtr expr2 = parseTerm();
      expr = std::make_shared<Binary>(op, expr, expr2);
   }
   return expr;
}

ExpressionPtr Parser::parseTerm() throw(Exception) {
   ExpressionPtr expr = parseFactor();
   while (getToken().symbol == Token::MUL ||
	 getToken().symbol == Token::DIV) {
      Binary::BinaryOp op;
      switch (getToken().symbol) {
	 case Token::MUL: op = mulop; break;
	 case Token::DIV: op = divop; break;
	 default: /* does not happen */ break;
      }
      nextToken();
      ExpressionPtr expr2 = parseFactor();
      expr = std::make_shared<Binary>(op, expr, expr2);
   }
   return expr;
}

ExpressionPtr Parser::parseFactor() throw(Exception) {
   ExpressionPtr expr;
   switch (getToken().symbol) {
      case Token::VALUE:
	 expr = std::make_shared<Primary>(getToken().value);
	 nextToken();
	 break;
      case Token::LPAREN:
	 nextToken();
	 expr = parseAssignment();
	 if (getToken().symbol != Token::RPAREN) {
	    throw Exception(getToken(), "')' expected");
	 }
	 nextToken();
	 break;
      case Token::VARIABLE:
	 expr = std::make_shared<Variable>(symtab, getToken().identifier);
	 nextToken();
	 break;
      default:
	 throw Exception(getToken(), "expression expected");
   }
   return expr;
}
